from typing import Dict
from typing import List
from typing import Optional

from clarin.sru.constants import SRUDiagnostics
from clarin.sru.constants import SRUParam
from clarin.sru.constants import SRUVersion
from clarin.sru.diagnostic import SRUDiagnosticList
from clarin.sru.queryparser import SRUQuery
from clarin.sru.queryparser import SRUQueryParser
from fcsql import QueryNode
from fcsql import QueryParser
from fcsql import QueryParserException

from clarin.sru.fcs.constants import FCSQueryType

# ---------------------------------------------------------------------------


class FCSQuery(SRUQuery[QueryNode]):
    @property
    def query_type(self) -> str:
        return FCSQueryType.FCS.value


class FCSQueryParser(SRUQueryParser[QueryNode]):
    """Default query parser to parse FCS-QL."""

    def __init__(self) -> None:
        super().__init__()
        self.parser = QueryParser()

    @property
    def query_type(self) -> str:
        return FCSQueryType.FCS

    @property
    def query_parameter_names(self) -> List[str]:
        return [SRUParam.QUERY.value]

    def supports_version(self, version: Optional[SRUVersion]) -> bool:
        if not version:
            raise TypeError("Argument version is invalid/None.")
        # FCS-QL is only supported by SRU 2.0
        return version >= SRUVersion.VERSION_2_0

    def parse_query(
        self,
        version: SRUVersion,
        parameters: Dict[str, str],
        diagnostics: SRUDiagnosticList,
    ) -> Optional[SRUQuery[QueryNode]]:
        raw_query = parameters.get(SRUParam.QUERY)
        if raw_query is None:
            diagnostics.add_diagnostic(
                SRUDiagnostics.GENERAL_SYSTEM_ERROR,
                message="no query passed to query parser",
            )
            return None

        try:
            parsed_query: QueryNode = self.parser.parse(raw_query)
            return FCSQuery(raw_query, parsed_query)
        except QueryParserException as ex:
            diagnostics.add_diagnostic(
                SRUDiagnostics.QUERY_SYNTAX_ERROR, message=str(ex)
            )
        except Exception:
            diagnostics.add_diagnostic(
                SRUDiagnostics.GENERAL_SYSTEM_ERROR,
                message="Unexpected error while parsing query.",
            )
        return None


# ---------------------------------------------------------------------------



import logging
from dataclasses import dataclass
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

import jwt
from clarin.sru.constants import SRUDiagnostics
from clarin.sru.exception import SRUConfigException
from clarin.sru.exception import SRUException
from clarin.sru.server.auth import SRUAuthenticationInfo
from clarin.sru.server.auth import SRUAuthenticationInfoProvider
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
from jwt.algorithms import RSAAlgorithm
from werkzeug import Request

# from typing import Self  # 3.11


# ---------------------------------------------------------------------------


LOGGER = logging.getLogger(__name__)


ALGORITHM_NAME = "RS256"


# ---------------------------------------------------------------------------


class AuthenticationInfo(SRUAuthenticationInfo):
    def __init__(self, subject: str):
        super().__init__()
        self._subject = subject

    @property
    def authentication_method(self) -> str:
        return "JWT"

    @property
    def subject(self) -> str:
        return self._subject


@dataclass(frozen=True)
class Key:
    key_id: str
    public_key: RSAPublicKey


@dataclass(frozen=True)
class Verifier:
    key_id: str
    public_key: RSAPublicKey
    jwt: jwt.PyJWT
    claims: Dict[str, Any]

    def decode(self, token: str, no_verification: bool = False) -> Dict[str, Any]:
        # TODO: we could probably bypass most checks by using internals like jwt.PyJWS()._load(token)
        if no_verification:
            return self.jwt.decode(token, options=dict(verify_signature=False))
        return self.jwt.decode(
            token,
            self.public_key,  # type: ignore
            algorithms=[ALGORITHM_NAME],
            **self.claims,
        )

    def verify(self, token: str) -> bool:
        # from `api_jws.py` -> `decode_complete`/`_load`/`_verify_signature`
        # TODO: not sure, better to use the stable JWT API instead of internals
        payload, signing_input, header, signature = jwt.PyJWS()._load(token)
        # algorithm = RSAAlgorithm(RSAAlgorithm.SHA256)
        # return algorithm.verify(signing_input, self.public_key, signature)
        try:
            self.jwt.decode(
                token,
                self.public_key,  # type: ignore
                algorithms=[ALGORITHM_NAME],
                **self.claims,
            )
        except (
            jwt.exceptions.InvalidAudienceError,
            jwt.exceptions.ExpiredSignatureError,
            # jwt.exceptions.InvalidIssuerError,  # NOTE: test cases required, but not 'issuers' not set, so not raised
            jwt.exceptions.InvalidIssuedAtError,  # NOTE: test cases required
        ):
            # e.g. 'Invalid audience'
            # e.g. 'Signature has expired'
            # TODO: what if verifiers with different leeways?
            raise
        except jwt.exceptions.MissingRequiredClaimError:
            # this is required, if missing, raise?
            # e.g. 'Token is missing the "aud" claim'
            raise
        except jwt.exceptions.ImmatureSignatureError:
            # TODO: fail on signature not yet 'live'?
            # e.g. 'The token is not yet valid (iat)'
            return False
        except (
            jwt.exceptions.InvalidSignatureError,  # is a DecodeError
            jwt.exceptions.DecodeError,
            jwt.exceptions.InvalidAlgorithmError,
        ):
            # here we would optimally try other keys as we could serve multiple clients
            # e.g. 'Signature verification failed'
            # e.g. 'Invalid crypto padding' / 'Invalid header string: ...' - some parts are missing
            return False
        return True


# ---------------------------------------------------------------------------


class AuthenticationProvider(SRUAuthenticationInfoProvider):
    def __init__(self, verifiers: List[Verifier]) -> None:
        super().__init__()
        self.verifiers = verifiers

    @property
    def key_count(self) -> int:
        if not self.verifiers:
            return 0
        return len(self.verifiers)

    # ----------------------------------------------------

    def get_AuthenticationInfo(
        self, request: Request
    ) -> Optional[SRUAuthenticationInfo]:
        value = request.headers.get("Authentication")
        if not value or value.isspace():
            return None

        LOGGER.debug("Found Authentication header with: '%s'", value)
        if not value.lower().startswith("bearer"):
            LOGGER.debug(
                "Authentication header with incorrect format. Expected start: 'Bearer '"
            )
            return None

        token = value[6:].strip()
        if not token:
            LOGGER.debug("No bearer token in Authentication header?")
            return None

        return self._check_token(token)

    def _check_token(self, token_raw: str) -> AuthenticationInfo:
        try:
            token = jwt.PyJWT().decode(token_raw, options=dict(verify_signature=False))
            LOGGER.debug(
                "Token: jti=%s, iss=%s, aud=%s, sub=%s, iat=%s, exp=%s, nbt=%s",
                token.get("jti"),
                token.get("iss"),
                token.get("aud"),
                token.get("sub"),
                token.get("iat"),
                token.get("exp"),
                token.get("nbt"),
            )

            # TODO: sanitize 'sub' of JWT token
            token["sub"] = str(token["sub"]) if "sub" in token else ""

            if not self.verifiers:
                LOGGER.warning("No JWT verifiers found. Return unverified 'sub' claim.")
                # NOTE: `token.get("sub")` should return a `str`
                return AuthenticationInfo(str(token.get("sub")))

            for verifier in self.verifiers:
                try:
                    LOGGER.debug(
                        "Trying to verify token with key '%s'", verifier.key_id
                    )
                    if verifier.verify(token_raw):
                        return AuthenticationInfo(str(token.get("sub")))
                # TODO: figure out exceptions we want to catch or ignore
                except jwt.exceptions.InvalidAudienceError as ex:
                    # java: InvalidClaimException
                    # e.g. audience does not match
                    raise SRUException(
                        SRUDiagnostics.AUTHENTICATION_ERROR,
                        "error processing request authentication",
                        message=str(ex),
                    ) from ex
                except jwt.exceptions.ExpiredSignatureError as ex:
                    raise SRUException(
                        SRUDiagnostics.AUTHENTICATION_ERROR,
                        "error processing request authentication",
                        message="token expired",
                    ) from ex

            raise SRUException(
                SRUDiagnostics.AUTHENTICATION_ERROR,
                "error processing request authentication",
                message="Could not verify JSON Web token signature.",
            )

        except jwt.PyJWTError as ex:
            raise SRUException(
                SRUDiagnostics.AUTHENTICATION_ERROR,
                "error processing request authentication",
                message="Could not decode JSON Web token",
            ) from ex

    # ----------------------------------------------------

    class Builder:
        def __init__(self):
            self.keys = List[Key]
            self.audiences = List[str]
            self.ignore_IssuedAt = False
            self.leeway_IssuedAt = -1
            self.leeway_ExpiresAt = -1
            self.leeway_NotBefore = -1

        @classmethod
        def create(cls) -> "AuthenticationProvider.Builder":
            return AuthenticationProvider.Builder()

        def with_audience(self, audience: str) -> "AuthenticationProvider.Builder":
            self.audiences.add(audience)
            return self

        def with_ignore_IssuedAt(self) -> "AuthenticationProvider.Builder":
            self.ignore_IssuedAt = True
            return self

        def with_IssuedAt(self, leeway: int) -> "AuthenticationProvider.Builder":
            if leeway < 0:
                raise ValueError("leeway < 0")
            self.leeway_IssuedAt = leeway
            return self

        def with_ExpiresAt(self, leeway: int) -> "AuthenticationProvider.Builder":
            if leeway < 0:
                raise ValueError("leeway < 0")
            self.leeway_ExpiresAt = leeway
            return self

        def with_NotBefore(self, leeway: int) -> "AuthenticationProvider.Builder":
            if leeway < 0:
                raise ValueError("leeway < 0")
            self.leeway_NotBefore = leeway
            return self

        def with_public_key(
            self, key_id: str, key: Any
        ) -> "AuthenticationProvider.Builder":
            self._load_public_key(key_id, key)
            return self

        def _load_public_key(self, key_id: str, key: Any) -> None:
            # RSASSA-PKCS1-v1_5 with SHA-256 ("RS256")
            # see: https://www.rfc-editor.org/rfc/rfc7519#section-8
            # https://pyjwt.readthedocs.io/en/stable/faq.html#how-can-i-extract-a-public-private-key-from-a-x509-certificate
            try:
                # algorithm_name = "RS256"
                algorithm = RSAAlgorithm(RSAAlgorithm.SHA256)
                key_obj: RSAPublicKey = algorithm.prepare_key(key)
                self.keys[key_id] = key_obj
            except (ValueError, TypeError) as ex:
                raise SRUConfigException(f"Failed to load key '{key_id}'") from ex

        def build(self) -> "AuthenticationProvider":
            verifiers: List[Verifier] = list()

            for key_id, key in self.keys.items():
                claims = dict()
                if self.audiences:
                    claims["audience"] = list(self.audiences)

                options = dict()
                if self.ignore_IssuedAt:
                    options["verify_iat"] = False
                else:
                    pass
                if self.leeway_ExpiresAt > 0:
                    options["leeway"] = self.leeway_ExpiresAt

                # both IssuedAd (IAT) and NotBefore (NBF) should be the
                # same based on fcs-simple-client (auth)
                # unfortunately PyJWT does not have separate leeways
                # see: com.auth0.jwt.interfaces.Verification
                # TODO: warn? or just use min/max/avg from all?
                # if self.leeway_IssuedAt > 0:
                #     options["leeway"] = self.leeway_IssuedAt
                # if self.leeway_NotBefore > 0:
                #     options["leeway"] = self.leeway_NotBefore

                verifiers.append(
                    Verifier(
                        key_id=key_id,
                        public_key=key,
                        jwt=jwt.PyJWT(**options),
                        claims=claims,
                    )
                )

            return AuthenticationProvider(verifiers)


# ---------------------------------------------------------------------------



"""Various useful constants for CLARIN-FCS endpoints."""
from enum import Enum

from clarin.sru.constants import SRUQueryType

# ---------------------------------------------------------------------------
# Diagnostics


FCS_DIAGNOSTIC_URI_PREFIX = "http://clarin.eu/fcs/diagnostic/"


class FCSDiagnostics(str, Enum):
    def __new__(cls, nr: int, fatal: bool, description: str):
        obj = str.__new__(cls, f"{FCS_DIAGNOSTIC_URI_PREFIX}{nr}")
        obj._value_ = f"{FCS_DIAGNOSTIC_URI_PREFIX}{nr}"
        obj.nr = nr
        obj.fatal = fatal
        obj.description = description
        obj.__doc__ = description
        return obj

    nr: int
    fatal: bool
    description: str

    # fmt: off

    PERSISTENT_IDENTIFIER_INVALID = (1, False, "Persistent identifier passed by the Client for restricting the search is invalid.")
    RESOURCE_TOO_LARGE_CONTEXT_ADJUSTED = (2, False, "Resource set too large. Query context automatically adjusted.")
    RESOURCE_TOO_LARGE_CANNOT_PERFORM_QUERY = (3, True, "Resource set too large. Cannot perform Query.")
    REQUESTED_DATA_VIEW_INVALID = (4, False, "Requested Data View not valid for this resource.")
    GENERAL_QUERY_SYNTAX_ERROR = (10, True, "General query syntax error.")
    GENERAL_QUERY_TOO_COMPLEX_CANNOT_PERFORM_QUERY = (11, True, "Query too complex. Cannot perform Query.")
    QUERY_WAS_REWRITTEN = (12, False, "Query was rewritten.")
    GENERAL_PROCESSING_HINT = (13, False, "General processing hint.")

    # fmt: on


# ---------------------------------------------------------------------------
# enums


class FCSQueryType(str, Enum):
    def __str__(self) -> str:
        return self.value

    FCS = "fcs"
    CQL = SRUQueryType.CQL
    SEARCH_TERMS = SRUQueryType.SEARCH_TERMS


class FCSLayerType(str, Enum):
    def __str__(self) -> str:
        return self.value

    TEXT = "text"
    """Textual representation of resource, also the layer that is used in Basic Search, String"""
    LEMMA = "lemma"
    """Lemmatisation, String"""
    POS = "pos"
    """Part-of-Speech annotations, Universal POS tags"""
    ORTH = "orth"
    """Orthographic transcription of (mostly) spoken resources, String"""
    NORM = "norm"
    """Orthographic normalization of (mostly) spoken resources, String"""
    PHONETIC = "phonetic"
    """Phonetic transcription, SAMPA"""


LAYER_TYPE_EXTENSION_PREFIX = "x-"


# ---------------------------------------------------------------------------
# namespaces


# from: https://docs.oracle.com/javase/7/docs/api/javax/xml/XMLConstants.html
# from xml.dom import XML_NAMESPACE
XML_NS_PREFIX = "xml"
XML_NS_URI = "http://www.w3.org/XML/1998/namespace"


FCS_NS = "http://clarin.eu/fcs/resource"
FCS_PREFIX = "fcs"
ED_NS = "http://clarin.eu/fcs/endpoint-description"
ED_PREFIX = "ed"

RI_NS_LEGACY = "http://clarin.eu/fcs/1.0/resource-info"


class FCSDataViewNamespaces(str, Enum):
    def __new__(cls, prefix: str, namespace: str, mimetype: str):
        obj = str.__new__(cls, namespace)
        obj._value_ = namespace
        obj.prefix = prefix
        obj.namespace = namespace
        obj.mimetype = mimetype
        return obj

    prefix: str
    namespace: str
    mimetype: str

    # fmt: off
    HITS = ("hits", "http://clarin.eu/fcs/dataview/hits", "application/x-clarin-fcs-hits+xml")
    KWIC = ("kwic", "http://clarin.eu/fcs/1.0/kwic", "application/x-clarin-fcs-kwic+xml")
    ADV = ("adv", "http://clarin.eu/fcs/dataview/advanced", "application/x-clarin-fcs-adv+xml")
    # fmt: on


# ---------------------------------------------------------------------------


class Capabilities(str, Enum):
    def __str__(self) -> str:
        return self.value

    BASIC_SEARCH = "http://clarin.eu/fcs/capability/basic-search"
    ADVANCED_SEARCH = "http://clarin.eu/fcs/capability/advanced-search"


# ---------------------------------------------------------------------------
# params?


LANG_EN = "en"

RESOURCE_URI_PREFIX = "resource:"

X_FCS_ENDPOINT_DESCRIPTION = "x-fcs-endpoint-description"
X_FCS_CONTEXT = "x-fcs-context"


# ---------------------------------------------------------------------------
# params


class FCSAuthenticationParam(str, Enum):
    def __str__(self) -> str:
        return self.value

    ENABLE = "eu.clarin.sru.server.fcs.authentication.enable"
    AUDIENCE = "eu.clarin.sru.server.fcs.authentication.audience"
    IGNORE_ISSUEDAT = "eu.clarin.sru.server.fcs.authentication.ignoreIssuedAt"
    ACCEPT_ISSUEDAT = "eu.clarin.sru.server.fcs.authentication.acceptIssuedAt"
    ACCEPT_EXPIRESAT = "eu.clarin.sru.server.fcs.authentication.acceptExpiresAt"
    ACCEPT_NOTBEFORE = "eu.clarin.sru.server.fcs.authentication.acceptNotBefore"

    # prefix
    PUBLIC_KEY_PREFIX = "eu.clarin.sru.server.fcs.authentication.key."


# ---------------------------------------------------------------------------



import importlib.resources
import logging
import warnings
from abc import ABCMeta
from abc import abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Dict
from typing import List
from typing import Optional
from xml.sax import ContentHandler
from xml.sax import SAXException

import cql
import cql.parser
from clarin.sru.constants import SRUDiagnostics
from clarin.sru.diagnostic import SRUDiagnosticList
from clarin.sru.exception import SRUConfigException
from clarin.sru.exception import SRUException
from clarin.sru.queryparser import SRUQueryParserRegistry
from clarin.sru.server.auth import SRUAuthenticationInfoProvider
from clarin.sru.server.auth import SRUAuthenticationInfoProviderFactory
from clarin.sru.server.config import SRUServerConfig
from clarin.sru.server.request import SRURequest
from clarin.sru.server.result import SRUExplainResult
from clarin.sru.server.result import SRUScanResultSet
from clarin.sru.server.server import SRUSearchEngine
from clarin.sru.xml.writer import SRUXMLStreamWriter
from clarin.sru.xml.writer import XMLStreamWriterHelper

from clarin.sru.fcs.constants import ED_NS
from clarin.sru.fcs.constants import ED_PREFIX
from clarin.sru.fcs.constants import RESOURCE_URI_PREFIX
from clarin.sru.fcs.constants import X_FCS_ENDPOINT_DESCRIPTION
from clarin.sru.fcs.constants import XML_NS_PREFIX
from clarin.sru.fcs.constants import XML_NS_URI
from clarin.sru.fcs.constants import FCSAuthenticationParam
from clarin.sru.fcs.queryparser import FCSQueryParser
from clarin.sru.fcs.server.auth import AuthenticationProvider

# ---------------------------------------------------------------------------


LOGGER = logging.getLogger(__name__)


# ---------------------------------------------------------------------------


@dataclass(frozen=True)
class Layer:
    """This class is used to information about a Layers that is
    available by the endpoint."""

    class ContentEncoding(str, Enum):
        """The content encoding policy for a Layer."""

        def __str__(self) -> str:
            return self.value

        VALUE = "value"
        """Value information is encoded as element content in this
        layer."""
        EMPTY = "empty"
        """No additional value information is encoded for this layer."""

    # ----------------------------------------------------

    id: str
    """The identifier of the layer"""
    result_id: str
    """The unique URI that used in the Advanced Data View to refer
    to this layer"""
    type: str
    """The type identifier for the layer"""
    encoding: ContentEncoding
    """The content encoding for this layer"""
    qualifier: Optional[str] = None
    """An optional layer qualifier to be used in FCS-QL to refer to
    this layer or ``None``. Defaults to ``None``."""
    alt_ValueInfo: Optional[str] = None
    """An additional information about the layer or ``None``.
    Defaults to ``None``."""
    alt_ValueInfo_url: Optional[str] = None
    """An additional URI for pointing to more information about the
    layer or ``None``. Defaults to ``None``."""

    def __post_init__(self):
        if self.id is None:
            raise TypeError("id is None")
        if self.result_id is None:
            raise TypeError("result_id is None")
        if self.type is None:
            raise TypeError("type is None")
        if self.encoding is None:
            raise TypeError("encoding is None")

    def __str__(self) -> str:
        return (
            f"{self.__class__.__name__}"
            f"[id={self.id}, result-id={self.result_id}, type={self.type}"
            f"{', qualifier=' + self.qualifier if self.qualifier else ''}]"
        )


# ---------------------------------------------------------------------------


@dataclass(frozen=True)
class DataView:
    """This class is used to hold information about a data view that
    is implemented by the endpoint."""

    class DeliveryPolicy(str, Enum):
        """Enumeration to indicate the delivery policy of a data view."""

        def __str__(self) -> str:
            return self.value

        SEND_BY_DEFAULT = "send-by-default"
        """The data view is sent automatically  by the endpoint."""
        NEED_TO_REQUEST = "need-to-request"
        """A client must explicitly request the endpoint."""

    # ----------------------------------------------------

    identifier: str
    """A unique short identifier for the data view"""
    mimetype: str
    """The MIME type of the data view"""
    deliveryPolicy: DeliveryPolicy
    """The delivery policy for this data view"""

    def __post_init__(self):
        if self.identifier is None:
            raise TypeError("identifier is None")
        elif self.identifier.isspace():
            raise ValueError("identifier is empty")
        if self.mimetype is None:
            raise TypeError("mimetype is None")
        elif self.mimetype.isspace():
            raise ValueError("mimetype is empty")
        if self.deliveryPolicy is None:
            raise TypeError("identifier is None")

    def __str__(self) -> str:
        return (
            f"{self.__class__.__name__}"
            f"[identifier={self.identifier}, mimeType={self.mimetype}]"
        )


# ---------------------------------------------------------------------------


@dataclass(frozen=True)
class ResourceInfo:
    """This class implements a resource info record, which provides
    supplementary information about a resource that is available at
    the endpoint."""

    pid: str
    """Rhe persistent identifier of the resource"""
    title: Dict[str, str]
    """The title of the resource represented as a map with pairs of
    language code and title"""
    description: Optional[Dict[str, str]]
    """The description of the resource represented as a map with pairs
    of language code and description or ``None`` if not applicable"""
    landing_page_uri: Optional[str]
    """A URI to the landing page of the resource or ``None`` if not
    applicable"""
    languages: List[str]
    """The languages represented within this resource represented as
    a list of ISO-632-3 three letter language codes"""
    available_DataViews: List[DataView]
    """The list of available data views for this resource"""
    available_Layers: Optional[List[Layer]] = None
    """The list if layers available for Advanced Search or ``None``
    if not applicable"""
    sub_Resources: Optional[List["ResourceInfo"]] = None
    """A list of resource sub-ordinate to this resource or ``None``
    if not applicable"""

    def __post_init__(self):
        if self.pid is None:
            raise TypeError("pid is None")
        if self.title is None:
            raise TypeError("title is None")
        elif not self.title:
            raise ValueError("title is empty")
        if self.languages is None:
            raise TypeError("languages is None")
        elif not self.languages:
            raise ValueError("languages is empty")
        if self.available_DataViews is None:
            raise TypeError("available_DataViews is None")

        # clear out
        if not self.description:
            object.__setattr__(self, "description", None)
        if not self.available_Layers:
            object.__setattr__(self, "available_Layers", None)
        if not self.sub_Resources:
            object.__setattr__(self, "sub_Resources", None)

    # ----------------------------------------------------

    def get_title(self, language: str) -> Optional[str]:
        """Get the title of the resource for a specific language code.

        Args:
            language: the language code (ISO-632-3 three letter
                language code)

        Returns:
            Optional[str]: the title for the language code or ``None`` if not title for this language code exists
        """
        return self.title.get(language)

    def get_description(self, language: str) -> Optional[str]:
        """Get the description of the resource for a specific language
        code.

        Args:
            language: the language code (ISO-632-3 three letter
                language code)

        Returns:
            Optional[str]: the description for the language code or ``None`` if not description for this language code exists.
        """
        if not self.description:
            return None
        return self.description.get(language)

    # ----------------------------------------------------

    def has_available_Layers(self) -> bool:
        """Check if any layers are available for Advanced Search.

        Returns:
            bool: ``True`` if any layer for Advanced Search is available, ``False`` otherwise
        """
        return bool(self.available_Layers)

    def has_sub_Resources(self) -> bool:
        """Determine, if this resource has sub-resources.

        Returns:
            bool: ``True`` if the resource has sub-resources, ``False`` otherwise
        """
        return bool(self.sub_Resources)


# ---------------------------------------------------------------------------


class EndpointDescription(metaclass=ABCMeta):
    """An interface for abstracting resource endpoint descriptions.
    This interface allows you to provide a version of a endpoint
    description tailored to your environment.

    The implementation of this interface **must** be thread-safe.
    """

    VERSION_1 = 1
    """Constant for endpoint description version number for FCS 1.0"""
    VERSION_2 = 2
    """Constant for endpoint description version number for FCS 2.0"""

    PID_ROOT = "root"
    """Constant for a (synthetic) persistent identifier identifying
    the top-most (= root) resources in the resource inventory."""

    @abstractmethod
    def destroy(self) -> None:
        """Destroy the resource info inventory. Use this method for
        any cleanup the resource info inventory needs to perform upon
        termination, i.e. closing of persistent database connections,
        etc."""

    @abstractmethod
    def get_version(self) -> int:
        """Get the version number of this endpoint description.
        Valid version are 1 for FCS 1.0 and 2 for FCS 2.0.

        Returns:
            int: the version number for this endpoint description
        """

    @abstractmethod
    def is_version(self, version: int) -> bool:
        """Check if this endpoint description is in a certain version.

        Args:
            version: the version to check for

        Returns:
            bool: ``True`` if version number matches
        """

    @abstractmethod
    def get_capabilities(self) -> List[str]:
        """Get the list of capabilities supported by this endpoint.
        The list contains the appropriate URIs defined by the
        CLARIN-FCS specification to indicate support for certain
        capabilities. This list **must** always contain at least
        ``http://clarin.eu/fcs/capability/basic-search`` for the
        **Basic Search** capability.

        The implementation of this method **must** be thread-safe.

        Returns:
            List[str]: the list of capabilities supported by this endpoint
        """

    @abstractmethod
    def get_supported_DataViews(self) -> List[DataView]:
        """Get the list of data views supported by this endpoint.
        This list **must** always contain an entry for the
        **Generic Hits (HITS)** data view.

        The implementation of this method **must** be thread-safe.

        Returns:
            List[DataView]: the list of data views supported by this endpoint
        """

    @abstractmethod
    def get_supported_Layers(self) -> List[Layer]:
        """Get the list of layers that are supported in Advanced
        Search by this endpoint.

        The implementation of this method **must** be thread-safe.

        Returns:
            List[Layer]: the list of layers supported in Advanced Search by this endpoint
        """

    @abstractmethod
    def get_ResourceInfos(self, pid: str) -> Optional[List[ResourceInfo]]:
        """Get a list of all resources sub-ordinate to a resource
        identified by a given persistent identifier.

        The implementation of this method **must** be thread-safe.

        Args:
            pid: the persistent identifier of the superior resource

        Returns:
            List[ResourceInfo]: a list of all sub-ordinate ResourceInfo or ``None`` if not applicable

        Raises:
            `SRUException`: if an error occurred
        """


class EndpointDescriptionBase(EndpointDescription, metaclass=ABCMeta):
    """An abstract base class for implementing endpoint descriptions.
    It already implements the methods required for capabilities and
    supported data views."""

    def __init__(
        self,
        version: int,
        capabilities: List[str],
        supported_DataViews: List[DataView],
        supported_Layers: Optional[List[Layer]],
    ) -> None:
        """[Constructor]

        Args:
            version: version of this endpoint description
            capabilities: a list of capabilities supported by this endpoint
            supported_DataViews: a list of data views that are supported by this endpoint
            supported_Layers: a list of layers that are supported by this endpoint

        Raises:
            TypeError: if arguments are invalid (None)
            ValueError: if argument values are not allowed
        """
        super().__init__()
        if version not in (1, 2):
            raise ValueError("version must be either 1 or 2")
        if capabilities is None:
            raise TypeError("capabilities is None")
        if not capabilities:
            raise ValueError("capabilities is empty")
        for capability in capabilities:
            if not capability:
                raise ValueError("capabilities must not contain a 'None'/empty item")
        if supported_DataViews is None:
            raise TypeError("supported_DataViews is None")
        if not supported_DataViews:
            raise ValueError("supported_DataViews is empty")
        for supported_DataView in supported_DataViews:
            if not supported_DataView:
                raise ValueError(
                    "supported_DataViews must not contain a 'None'/empty item"
                )
        if not supported_Layers:
            supported_Layers = list()
        for supported_Layer in supported_Layers:
            if not supported_Layer:
                raise ValueError(
                    "supported_Layers must not contain a 'None'/empty item"
                )

        self.version = version
        self.capabilities = list(capabilities)
        self.supported_DataViews = list(supported_DataViews)
        self.supported_Layers = list(supported_Layers)

    def get_version(self) -> int:
        return self.version

    def is_version(self, version: int) -> bool:
        return self.version == version

    def get_capabilities(self) -> List[str]:
        return self.capabilities

    def get_supported_DataViews(self) -> List[DataView]:
        return self.supported_DataViews

    def get_supported_Layers(self) -> List[Layer]:
        return self.supported_Layers


class SimpleEndpointDescription(EndpointDescriptionBase):
    """A very simple implementation of an endpoint description that
    is initialized from static information supplied at construction
    time. Mostly used together with `SimpleEndpointDescriptionParser`,
    but it is agnostic how the static list of resource info records
    is generated."""

    def __init__(
        self,
        version: int,
        capabilities: List[str],
        supported_DataViews: List[DataView],
        supported_Layers: List[Layer],
        resources: List[ResourceInfo],
        pid_case_sensitive: bool,
    ) -> None:
        """Constructor.

        Args:
            version: version of this endpoint description
            capabilities: a list of capabilities supported by this endpoint
            supported_DataViews: a list of data views that are supported by this endpoint
            supported_Layers: a list of layers supported for Advanced Search by this endpoint or ``None``
            resources: a static list of resource info records
            pid_case_sensitive: ``True`` if comparison of persistent identifiers should be performed case-sensitive, ``False`` otherwise

        Raises:
            TypeError: if resources are None
        """
        super().__init__(version, capabilities, supported_DataViews, supported_Layers)
        if resources is None:
            raise TypeError("entries/resources is None")

        self.entries = list(resources)
        self.pid_case_sensitive = pid_case_sensitive

    def destroy(self) -> None:
        pass

    def get_ResourceInfos(self, pid: str) -> Optional[List[ResourceInfo]]:
        if pid is None:
            raise TypeError("pid is None")
        if pid.isspace():
            raise ValueError("pid is empty")

        if not self.pid_case_sensitive:
            pid = pid.lower()

        if pid == EndpointDescription.PID_ROOT:
            return self.entries
        else:
            ri = self.find_recursive(self.entries, pid)
            if ri:
                return ri.sub_Resources
        return None

    def find_recursive(
        self, items: Optional[List[ResourceInfo]], pid: str
    ) -> Optional[ResourceInfo]:
        if items:
            for item in items:
                if self.pid_case_sensitive:
                    if pid == item.pid:
                        return item
                else:
                    if pid.lower() == item.pid.lower():
                        return item
                if item.has_sub_Resources():
                    ri = self.find_recursive(item.sub_Resources, pid)
                    if ri:
                        return ri
        return None


# ---------------------------------------------------------------------------


class SimpleEndpointSearchEngineBase(
    SRUAuthenticationInfoProviderFactory, SRUSearchEngine, metaclass=ABCMeta
):
    """A base class for implementing a simple search engine to be
    used as a CLARIN-FCS endpoint."""

    def __init__(self) -> None:
        super().__init__()
        self.endpoint_description: Optional[EndpointDescription] = None

    # ----------------------------------------------------
    # non-overwritable (you really shouldn't)

    def init(
        self,
        config: SRUServerConfig,
        query_parser_registry_builder: SRUQueryParserRegistry.Builder,
        params: Dict[str, str],
    ) -> None:
        """This method should not be overridden. Perform your custom
        initialization in the `do_init` method instead.

        See also:
            `do_init`,
            `SRUSearchEngine.init`
        """
        LOGGER.debug("Initializing")
        super().init(config, query_parser_registry_builder, params)

        query_parser_registry_builder.register(FCSQueryParser())

        LOGGER.debug("Initializing search engine implementation")
        self.do_init(config, query_parser_registry_builder, params)

        LOGGER.debug("Initizalizing endpoint description")
        self.endpoint_description = self.create_EndpointDescription(
            config, query_parser_registry_builder, params
        )
        if not self.endpoint_description:
            LOGGER.error(
                "SimpleEndpointSearchEngineBase implementation error: "
                "create_EndpointDescription() returned None"
            )
            raise SRUConfigException(
                "create_EndpointDescription() returned no valid"
                " implementation of an EndpointDescription"
            )

    def destroy(self) -> None:
        """This method should not be overridden. Perform you custom
        cleanup in the `do_destroy` method.

        See also:
            `do_destroy`,
            `SRUSearchEngine.destroy`
        """
        LOGGER.debug("Performing cleanup of endpoint description")
        if self.endpoint_description:
            self.endpoint_description.destroy()

        LOGGER.debug("Performing cleanup of search engine")
        self.do_destroy()

        return super().destroy()

    def create_SRUAuthenticationInfoProvider(
        self, params: Dict[str, str]
    ) -> Optional[SRUAuthenticationInfoProvider]:
        enabled_str = params.get(FCSAuthenticationParam.ENABLE)
        if not enabled_str:
            return None
        enabled = SimpleEndpointSearchEngineBase._parse_bool(enabled_str)
        if not enabled:
            LOGGER.debug("Explictly disable authentication")
            return None

        LOGGER.debug("Enabling authentication")
        builder = AuthenticationProvider.Builder.create()

        audience = params.get(FCSAuthenticationParam.AUDIENCE)
        if audience:
            values = [v.strip() for v in audience.split(",") if v.strip()]
            if values:
                for value in values:
                    LOGGER.debug("Adding audience: %s", value)
                    builder.with_audience(value)

        ignore_IssuedAt = SimpleEndpointSearchEngineBase._parse_bool(
            params.get(FCSAuthenticationParam.IGNORE_ISSUEDAT)
        )
        if ignore_IssuedAt:
            LOGGER.debug("Will not verify 'iat' claim")
            builder.with_ignore_IssuedAt()
        else:
            leeway = SimpleEndpointSearchEngineBase._parse_int(
                params.get(FCSAuthenticationParam.ACCEPT_ISSUEDAT), -1
            )
            if leeway > 0:
                LOGGER.debug("Allowing %s seconds leeway for 'iat' claim", leeway)
                builder.with_IssuedAt(leeway)

        leeway = SimpleEndpointSearchEngineBase._parse_int(
            params.get(FCSAuthenticationParam.ACCEPT_EXPIRESAT), -1
        )
        if leeway > 0:
            LOGGER.debug("Allowing %s seconds leeway for 'exp' claim", leeway)
            builder.with_ExpiresAt(leeway)

        leeway = SimpleEndpointSearchEngineBase._parse_int(
            params.get(FCSAuthenticationParam.ACCEPT_NOTBEFORE), -1
        )
        if leeway > 0:
            LOGGER.debug("Allowing %s seconds leeway for 'nbf' claim", leeway)
            builder.with_NotBefore(leeway)

        # load keys
        for name, value in params.items():
            if not name.startswith(FCSAuthenticationParam.PUBLIC_KEY_PREFIX):
                continue
            key_id = name[len(FCSAuthenticationParam.PUBLIC_KEY_PREFIX.value) :].strip()
            if not key_id:
                raise SRUConfigException(
                    f"init-parameter: '{name}' is invalid: key_id is empty!"
                )
            # key_filename = value
            LOGGER.debug("key_id = %s, key_file = %s", key_id, value)
            if value.startswith(RESOURCE_URI_PREFIX):
                LOGGER.debug("Loading key '%s' from resource '%s'", key_id, value)
                key = self._load_key_from_resource(value[len(RESOURCE_URI_PREFIX) :])
                builder.with_public_key(key_id, key)
            else:
                LOGGER.debug("Loading key '%s' from file '%s'", key_id, value)
                # NOTE: value is a filepath
                builder.with_public_key(key_id, value)

        auth_provider = builder.build()
        if auth_provider.key_count == 0:
            LOGGER.warning(
                "No keys configured, all well-formed tokens will be"
                " accepted. Make sure, youn know what you are doing!"
            )
        return auth_provider

    def _load_key_from_resource(self, path: str) -> bytes:
        package, name = path.split(":")

        if not importlib.resources.is_resource(package, name):
            raise SRUConfigException(f"Cannot open '{name}' in '{package}'")

        with importlib.resources.open_binary(package, name) as fp:
            return fp.read()

    # ----------------------------------------------------

    def explain(
        self,
        config: SRUServerConfig,
        request: SRURequest,
        diagnostics: SRUDiagnosticList,
    ) -> Optional[SRUExplainResult]:
        val = request.get_extra_request_data(X_FCS_ENDPOINT_DESCRIPTION)
        provide_epdesc = SimpleEndpointSearchEngineBase._parse_bool(val)

        if provide_epdesc and self.endpoint_description:

            class SRUExplainResultWithEndpointDescription(SRUExplainResult):
                def __init__(
                    self,
                    diagnostics: SRUDiagnosticList,
                    endpoint_description: EndpointDescription,
                ) -> None:
                    super().__init__(diagnostics)
                    self.endpoint_description = endpoint_description

                @property
                def has_extra_response_data(self) -> bool:
                    return True

                def write_extra_response_data(self, writer: SRUXMLStreamWriter) -> None:
                    SimpleEndpointSearchEngineBase._write_EndpointDescription(
                        writer, self.endpoint_description
                    )

            return SRUExplainResultWithEndpointDescription(
                diagnostics, self.endpoint_description
            )

        return None

    def scan(
        self,
        config: SRUServerConfig,
        request: SRURequest,
        diagnostics: SRUDiagnosticList,
    ) -> Optional[SRUScanResultSet]:
        """Handle a **scan** operation. This implementation provides
        support to CLARIN FCS resource enumeration. If you want to
        provide custom scan behavior for a different index, override
        the `do_scan` method.
        """
        return self.do_scan(config, request, diagnostics)

    # ----------------------------------------------------
    # overwritable (can/should be overwritten)

    @abstractmethod
    def create_EndpointDescription(
        self,
        config: SRUServerConfig,
        query_parser_registry_builder: SRUQueryParserRegistry.Builder,
        params: Dict[str, str],
    ) -> EndpointDescription:
        pass

    @abstractmethod
    def do_init(
        self,
        config: SRUServerConfig,
        query_parser_registry_builder: SRUQueryParserRegistry.Builder,
        params: Dict[str, str],
    ) -> None:
        """Initialize the search engine. This initialization should
        be tailed towards your environment and needs.

        Args:
            config: the `SRUServerConfig` object for this search engine
            query_parser_registry_builder: the `SRUQueryParserRegistry.Builder`
                object to be used for this search engine. Use to register
                additional query parsers with the `SRUServer`
            params: additional parameters from the server configuration

        Raises:
            SRUConfigException: if an error occurred
        """

    def do_destroy(self) -> None:
        """Destroy the search engine. Override this method for any
        cleanup the search engine needs to perform upon termination."""

    def do_scan(
        self,
        config: SRUServerConfig,
        request: SRURequest,
        diagnostics: SRUDiagnosticList,
    ) -> SRUScanResultSet:
        """[Deprecated]
        Handle a **scan** operation. The default implementation is
        a no-op. Override this method, if you want to provide a
        custom behavior.

        Args:
            config: the `SRUEndpointConfig` object that contains the
                endpoint configuration
            request: the `SRURequest` object that contains the
                request made to the endpoint
            diagnostics: the `SRUDiagnosticList` object for storing
                non-fatal diagnostics

        Returns:
            SRUScanResultSet: a `SRUScanResultSet` object or ``None``
                if this operation is not supported by this search
                engine

        Raises:
            `SRUException`: if an fatal error occurred
        """
        warnings.warn(
            "'do_scan' is deprecated. See Java implementation for more details?!",
            category=DeprecationWarning,
            stacklevel=2,
        )

        # not really sure what this does exactly, based on java implementation
        scan_clause = request.get_scan_clause()
        if not scan_clause:
            # NOTE: this should not happen?
            raise SRUException(
                SRUDiagnostics.GENERAL_SYSTEM_ERROR,
                message="Missing scan clause?",
            )
        elif (
            isinstance(scan_clause.root, cql.parser.CQLSearchClause)
            and scan_clause.root.index
        ):
            # CQLTermNode ?
            index = scan_clause.root.index.name
            raise SRUException(
                SRUDiagnostics.UNSUPPORTED_INDEX,
                index,
                message=f"scan operation on index '{index}' is not supported",
            )
        else:
            raise SRUException(
                SRUDiagnostics.QUERY_FEATURE_UNSUPPORTED,
                message="Scan clause too complex.",
            )

    # ----------------------------------------------------

    @staticmethod
    def _write_EndpointDescription(writer: ContentHandler, epdesc: EndpointDescription):
        writer = XMLStreamWriterHelper(writer)

        writer.startPrefixMapping(ED_PREFIX, ED_NS)
        writer.startElementNS(
            (ED_NS, "EndpointDescription"),
            attrs={"version": str(epdesc.get_version())},
        )

        # Capabilities
        with writer.element("Capabilities", ED_NS):
            for capability in epdesc.get_capabilities():
                with writer.element("Capability", ED_NS):
                    writer.characters(capability)

        # SupportedDataViews
        with writer.element("SupportedDataViews", ED_NS):
            for dataview in epdesc.get_supported_DataViews():
                if (
                    not dataview.deliveryPolicy
                    or dataview.deliveryPolicy not in DataView.DeliveryPolicy
                ):
                    raise SAXException(
                        f"invalid value for payload delivery policy: {dataview.deliveryPolicy}"
                    )
                with writer.element(
                    "SupportedDataView",
                    ED_NS,
                    attrs={
                        "id": dataview.identifier,
                        "delivery-policy": dataview.deliveryPolicy.value,
                    },
                ):
                    writer.characters(dataview.mimetype)

        # SupportedLayers (FCS 2.0)
        if epdesc.is_version(EndpointDescription.VERSION_2):
            if epdesc.get_supported_Layers():
                with writer.element("SupportedLayers", ED_NS):
                    for layer in epdesc.get_supported_Layers():
                        attrs = {"id": layer.id, "result-id": layer.result_id}
                        if layer.encoding == Layer.ContentEncoding.EMPTY:
                            attrs["type"] = "empty"
                        if layer.qualifier:
                            attrs["qualifier"] = layer.qualifier
                        if layer.alt_ValueInfo:
                            attrs["alt-value-info"] = layer.alt_ValueInfo
                            if layer.alt_ValueInfo_url:
                                attrs["alt-value-info-uri"] = layer.alt_ValueInfo_url
                        with writer.element("SupportedLayer", ED_NS, attrs=attrs):
                            writer.characters(layer.type)

        # Resources
        try:
            resources = epdesc.get_ResourceInfos(EndpointDescription.PID_ROOT)
            if not resources:
                raise SRUException("top level must contain resources")
            write_layers = epdesc.is_version(EndpointDescription.VERSION_2)
            SimpleEndpointSearchEngineBase._write_ResourceInfos(
                writer, resources, write_layers
            )
        except SRUException as ex:
            raise SAXException("error retriving top-level resources") from ex

        writer.endElementNS((ED_NS, "EndpointDescription"), None)
        writer.endPrefixMapping(ED_PREFIX)

    @staticmethod
    def _write_ResourceInfos(
        writer: ContentHandler, resources: List[ResourceInfo], write_layers: bool = True
    ):
        if resources is None:
            raise TypeError("resources is None")
        if not resources:
            return

        writer = XMLStreamWriterHelper(writer)

        # TODO: XML NS required or auto included/known?
        writer.startPrefixMapping(XML_NS_PREFIX, XML_NS_URI)
        writer.startElementNS((ED_NS, "Resources"))

        for resource in resources:
            writer.startElementNS((ED_NS, "Resource"), attrs={"pid": resource.pid})

            # title
            for lang, title in resource.title.items():
                with writer.element("Title", ED_NS, attrs={(XML_NS_URI, "lang"): lang}):
                    writer.characters(title)

            # description
            if resource.description:
                for lang, desc in resource.description.items():
                    with writer.element(
                        "Description", ED_NS, attrs={(XML_NS_URI, "lang"): lang}
                    ):
                        writer.characters(desc)

            # landing page
            if resource.landing_page_uri:
                with writer.element("LandingPageURI", ED_NS):
                    writer.characters(resource.landing_page_uri)

            # languages
            with writer.element("Languages", ED_NS):
                for language in resource.languages:
                    with writer.element("Language", ED_NS):
                        writer.characters(language)

            # available data views
            dvref = " ".join(dv.identifier for dv in resource.available_DataViews)
            writer.startElementNS((ED_NS, "AvailableDataViews"), attrs={"ref": dvref})
            writer.endElementNS((ED_NS, "AvailableDataViews"))

            # available layer (FCS 2.0)
            if write_layers and resource.available_Layers:
                lref = " ".join(ly.id for ly in resource.available_Layers)
                writer.startElementNS((ED_NS, "AvailableLayers"), attrs={"ref": lref})
                writer.endElementNS((ED_NS, "AvailableLayers"))

            # child resources
            subs = resource.sub_Resources
            if subs:
                SimpleEndpointSearchEngineBase._write_ResourceInfos(writer, subs)

            writer.endElementNS((ED_NS, "Resource"), None)

        writer.endElementNS((ED_NS, "Resources"), None)

    # ----------------------------------------------------

    @staticmethod
    def _parse_bool(val: Optional[str]) -> bool:
        """Convince method for parsing a string to boolean. Values
        ``1``, ``true`` and ``yes`` yield a ``True`` boolean value
        as a result, all others (include ``None``) a ``False``
        boolean value.

        Args:
            val: the string to parse

        Returns:
            bool: ``True`` if the supplied string was considered something representing a ``True`` boolean value, ``False`` otherwise
        """
        if not val:
            return False
        val = val.strip().lower()
        if not val:
            return False
        return val in ("true", "1", "yes")

    @staticmethod
    def _parse_int(val: Optional[str], default: int) -> int:
        if not val or val.isspace():
            return default
        try:
            return int(val)
        except ValueError:
            raise SRUConfigException("invalid long value")


# ---------------------------------------------------------------------------



import io
import logging
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import TextIO
from typing import Union
from urllib.parse import urlparse

from clarin.sru.exception import SRUConfigException
from lxml import etree

from clarin.sru.fcs.constants import ED_NS
from clarin.sru.fcs.constants import ED_PREFIX
from clarin.sru.fcs.constants import LANG_EN
from clarin.sru.fcs.constants import LAYER_TYPE_EXTENSION_PREFIX
from clarin.sru.fcs.constants import RI_NS_LEGACY
from clarin.sru.fcs.constants import XML_NS_URI
from clarin.sru.fcs.constants import Capabilities
from clarin.sru.fcs.constants import FCSDataViewNamespaces
from clarin.sru.fcs.constants import FCSLayerType
from clarin.sru.fcs.server.search import DataView
from clarin.sru.fcs.server.search import EndpointDescription
from clarin.sru.fcs.server.search import Layer
from clarin.sru.fcs.server.search import ResourceInfo
from clarin.sru.fcs.server.search import SimpleEndpointDescription

# ---------------------------------------------------------------------------


LOGGER = logging.getLogger(__name__)


# ---------------------------------------------------------------------------


class SimpleEndpointDescriptionParser:
    """A parser, that parses an XML file and produces a endpoint
    description with static list of resource info records. The XML
    file has the same format as the result format defined for
    endpoint description of the CLARIN-FCS specification. The `parse`
    returns a `SimpleEndpointDescription` instance."""

    @staticmethod
    def parse(url: Union[str, TextIO]) -> EndpointDescription:
        """Parse an XML file and return a static list of resource
        info records.

        Args:
            url: the URI pointing to the file to be parsed

        Returns:
            EndpointDescription: an `EndpointDescription` instance

        Raises:
            SRUConfigException: if an error occurred
            TypeError: if url is None
        """
        if url is None:
            raise TypeError("url is None")

        LOGGER.debug("Parsing endpoint description from: %s", url)

        try:
            parser = etree.XMLParser(
                ns_clean=False, remove_comments=True, strip_cdata=True
            )
            ed_doc: etree._ElementTree = etree.parse(url, parser)

            # TODO: validate with schema? (like fcs-sru-server config?)

            # Detect for deprecated resource-info catalog files and bail, if necessary
            # url.name if isinstance(url, io.TextIOWrapper) else url ?
            SimpleEndpointDescriptionParser._check_legacy_mode(ed_doc, url)

            # Parse on and create endpoint description ...
            return SimpleEndpointDescriptionParser._parse_EndpointDescription(ed_doc)
        except etree.XPathEvalError as ex:
            raise SRUConfigException("internal error") from ex
        except etree.XMLSyntaxError as ex:
            raise SRUConfigException("parsing error") from ex
        except SRUConfigException:
            raise
        except OSError as ex:
            raise SRUConfigException("error reading file") from ex
        except Exception as ex:
            raise SRUConfigException("internal error") from ex
        # lxml ParserConfigurationException ?

    # ----------------------------------------------------

    @staticmethod
    def _parse_EndpointDescription(doc: etree._ElementTree) -> EndpointDescription:
        version = SimpleEndpointDescriptionParser._parse_version(doc)
        LOGGER.debug("Endpoint description version is %s", version)

        capabilities = SimpleEndpointDescriptionParser._parse_Capabilities(doc)
        LOGGER.debug("CAPS: %s", capabilities)
        SimpleEndpointDescriptionParser._check_Capabilities(capabilities, version)

        xml_ids: Set[str] = set()  # used to check for uniqueness of id attribute
        supported_DataViews = SimpleEndpointDescriptionParser._parse_DataViews(
            doc, xml_ids
        )
        LOGGER.debug("DV: %s", supported_DataViews)
        SimpleEndpointDescriptionParser._check_DataViews(
            supported_DataViews, capabilities
        )

        supported_Layers = SimpleEndpointDescriptionParser._parse_Layers(doc, xml_ids)
        LOGGER.debug("L: %s", supported_Layers)
        SimpleEndpointDescriptionParser._check_Layers(supported_Layers, capabilities)

        resources = SimpleEndpointDescriptionParser._parse_ResourceInfos(
            doc, supported_DataViews, supported_Layers, version
        )
        if not resources:
            raise SRUConfigException(
                "No resources where defined in endpoint description"
            )
        if LOGGER.isEnabledFor(logging.DEBUG):
            LOGGER.debug("Dumping ResourceInfo:")
            SimpleEndpointDescriptionParser._dump_ResourceInfo(resources, 1)

        return SimpleEndpointDescription(
            version,
            capabilities,
            supported_DataViews,
            supported_Layers,
            resources,
            False,
        )

    # ----------------------------------------------------

    @staticmethod
    def _parse_version(doc: etree._ElementTree) -> int:
        version = -1
        nodes = doc.xpath(
            "//ed:EndpointDescription/@version", namespaces={ED_PREFIX: ED_NS}
        )
        if nodes:
            try:
                version = int(nodes[0])
                if version not in (1, 2):
                    raise SRUConfigException(
                        "Attribute @version element <EndpointDescription> must have a value of either '1' or '2'"
                    )
            except ValueError as ex:
                raise SRUConfigException("Cannot parse version number") from ex
        if version == -1:
            raise SRUConfigException(
                "Attribute @version missing on element <EndpointDescription>"
            )
        return version

    @staticmethod
    def _parse_Capabilities(doc: etree._ElementTree) -> List[str]:
        capabilities: List[str] = list()
        nodes = doc.xpath(
            "//ed:Capabilities/ed:Capability", namespaces={ED_PREFIX: ED_NS}
        )
        if nodes:
            LOGGER.debug("Parsing capabilities")
            for node in nodes:
                uri = node.text.strip()
                try:
                    urlparse(uri)
                except Exception:
                    raise SRUConfigException(
                        f"capability is not encoded as a proper URI: {uri}"
                    )
                if uri not in capabilities:
                    capabilities.append(uri)
                else:
                    LOGGER.warning("Ignoring duplicate capability entry for '%s'.", uri)

        else:
            LOGGER.warning("No capabilities where defined in endpoint configuration.")

        return capabilities

    @staticmethod
    def _check_Capabilities(capabilities: List[str], version: int):
        if Capabilities.BASIC_SEARCH not in capabilities:
            LOGGER.warning(
                "Capability '%s' was not defined in endpoint description;"
                " it was added to meet the specification. Please update your"
                " endpoint description!",
                Capabilities.BASIC_SEARCH,
            )
            capabilities.append(Capabilities.BASIC_SEARCH)

        if Capabilities.ADVANCED_SEARCH in capabilities and version < 2:
            LOGGER.warning(
                "Endpoint description is declared as version FCS 1.0 (@version = 1),"
                " but contains support for Advanced Search in capabilities list!"
                " FCS 1.0 only supports Basic Search."
            )

    @staticmethod
    def _parse_DataViews(doc: etree._ElementTree, xml_ids: Set[str]) -> List[DataView]:
        nodes = doc.xpath(
            "//ed:SupportedDataViews/ed:SupportedDataView",
            namespaces={ED_PREFIX: ED_NS},
        )
        if not nodes:
            LOGGER.error(
                "Endpoint configuration contains no valid information about supported data views"
            )
            raise SRUConfigException(
                "Endpoint configuration contains no valid information about supported data views"
            )

        dataviews: List[DataView] = list()
        LOGGER.debug("Parsing supported data views")
        for node in nodes:
            id = SimpleEndpointDescriptionParser._get_attribute(node, "id")
            if id is None:
                raise SRUConfigException(
                    "Element <SupportedDataView> must have a proper 'id' attribute"
                )

            if id in xml_ids:
                raise SRUConfigException(
                    f"The value of attribute 'id' of element <SupportedDataView> must be unique: {id}"
                )
            xml_ids.add(id)

            # TODO: refactor as constants?
            pval = SimpleEndpointDescriptionParser._get_attribute(
                node, "delivery-policy"
            )
            if pval is None:
                raise SRUConfigException(
                    "Element <SupportedDataView> must have a 'delivery-policy' attribute"
                )
            policy: DataView.DeliveryPolicy
            if DataView.DeliveryPolicy.SEND_BY_DEFAULT == pval:
                policy = DataView.DeliveryPolicy.SEND_BY_DEFAULT
            elif DataView.DeliveryPolicy.NEED_TO_REQUEST == pval:
                policy = DataView.DeliveryPolicy.NEED_TO_REQUEST
            else:
                raise SRUConfigException(
                    f"Invalid value '{pval}' for attribute 'delivery-policy' on element <SupportedDataView>"
                )

            mimetype: Optional[str] = node.text
            if mimetype:
                mimetype = mimetype.strip()
                if not mimetype:
                    mimetype = None
            if not mimetype:
                raise SRUConfigException(
                    "Element <SupportedDataView> must contain a MIME-type as content"
                )

            # check for duplicate entries ...
            for dataview in dataviews:
                if id == dataview.identifier:
                    raise SRUConfigException(
                        f"A <SupportedDataView> with the id '{id}' is already defined!"
                    )
                if mimetype == dataview.mimetype:
                    raise SRUConfigException(
                        f"A <SupportedDataView> with the MIME-type '{mimetype}' is already defined!"
                    )

            dataviews.append(
                DataView(identifier=id, mimetype=mimetype, deliveryPolicy=policy)
            )

        return dataviews

    @staticmethod
    def _check_DataViews(dataviews: List[DataView], capabilities: List[str]):
        # sanity check on data views
        has_HITS_view = has_ADV_view = False
        for dataview in dataviews:
            if dataview.mimetype == FCSDataViewNamespaces.HITS.mimetype:
                has_HITS_view = True
            elif dataview.mimetype == FCSDataViewNamespaces.ADV.mimetype:
                has_ADV_view = True
        if not has_HITS_view:
            raise SRUConfigException(
                f"Generic Hits Data View ({FCSDataViewNamespaces.HITS.mimetype})"
                " was not declared in <SupportedDataViews>"
            )
        if Capabilities.ADVANCED_SEARCH in capabilities and not has_ADV_view:
            raise SRUConfigException(
                "Endpoint claimes to support Advanced FCS but does not declare"
                f" Advanced Data View ({FCSDataViewNamespaces.ADV.mimetype})"
                "in <SupportedDataViews>"
            )

    @staticmethod
    def _parse_Layers(doc: etree._ElementTree, xml_ids: Set[str]) -> List[Layer]:
        layers: List[Layer] = list()
        nodes = doc.xpath(
            "//ed:SupportedLayers/ed:SupportedLayer",
            namespaces={ED_PREFIX: ED_NS},
        )
        if nodes:
            LOGGER.debug("Parsing supported layers")
            for node in nodes:
                id = SimpleEndpointDescriptionParser._get_attribute(node, "id")
                if id is None:
                    raise SRUConfigException(
                        "Element <SupportedLayer> must have a proper 'id' attribute"
                    )

                if id in xml_ids:
                    raise SRUConfigException(
                        f"The value of attribute 'id' of element <SupportedLayer> must be unique: {id}"
                    )
                xml_ids.add(id)

                # TODO: refactor as constants?
                result_id = SimpleEndpointDescriptionParser._get_attribute(
                    node, "result-id"
                )
                if result_id is None:
                    raise SRUConfigException(
                        "Element <SupportedLayer> must have a 'result-id' attribute"
                    )
                try:
                    urlparse(result_id)
                except Exception:
                    raise SRUConfigException(
                        "Attribute 'result-id' on Element <SupportedLayer>"
                        f" is not encoded as proper URI: {result_id}"
                    )

                rtype = SimpleEndpointDescriptionParser._clean_str(node.text)
                if not rtype:
                    raise SRUConfigException(
                        "Element <SupportedLayer> does not define a proper layer type"
                    )
                # sanity check on layer types
                if rtype not in [
                    lt.value for lt in FCSLayerType
                ] and not rtype.startswith(LAYER_TYPE_EXTENSION_PREFIX):
                    LOGGER.warning(
                        "layer type '%s' is not defined by specification", rtype
                    )

                qualifier = SimpleEndpointDescriptionParser._get_attribute(
                    node, "qualifier"
                )

                encoding = Layer.ContentEncoding.VALUE
                eval = SimpleEndpointDescriptionParser._get_attribute(node, "type")
                if eval:
                    if Layer.ContentEncoding.VALUE == eval:
                        encoding = Layer.ContentEncoding.VALUE
                    elif Layer.ContentEncoding.EMPTY == eval:
                        encoding = Layer.ContentEncoding.EMPTY
                    else:
                        raise SRUConfigException(f"Invalid layer encoding: {eval}")

                alt_value_info = SimpleEndpointDescriptionParser._get_attribute(
                    node, "alt-value-info"
                )
                alt_value_info_uri: Optional[str] = None
                if alt_value_info:
                    alt_value_info_uri = SimpleEndpointDescriptionParser._get_attribute(
                        node, "alt-value-info-uri"
                    )
                    try:
                        urlparse(alt_value_info_uri)
                    except Exception:
                        raise SRUConfigException(
                            "Attribute 'alt-value-info-uri' on Element <SupportedLayer>"
                            f" is not encoded as proper URI: {alt_value_info_uri}"
                        )

                layers.append(
                    Layer(
                        id=id,
                        result_id=result_id,
                        type=rtype,
                        encoding=encoding,
                        qualifier=qualifier,
                        alt_ValueInfo=alt_value_info,
                        alt_ValueInfo_url=alt_value_info_uri,
                    )
                )

        return layers

    @staticmethod
    def _check_Layers(layers: List[Layer], capabilities: List[str]):
        if layers and Capabilities.ADVANCED_SEARCH not in capabilities:
            LOGGER.warning(
                "Endpoint description has <SupportedLayer> but does not indicate"
                " support for Advanced Search. Please consider adding capability"
                " (%s) to your endpoint description to make use of layers!",
                Capabilities.ADVANCED_SEARCH,
            )

    @staticmethod
    def _parse_ResourceInfos(
        doc: etree._ElementTree,
        supported_DataViews: List[DataView],
        supported_Layers: List[Layer],
        version: int,
    ) -> List[ResourceInfo]:
        pids: Set[str] = set()
        has_ADV_view = any(
            dataview.mimetype == FCSDataViewNamespaces.ADV.mimetype
            for dataview in supported_DataViews
        )

        def _parse_resources(nodes) -> List[ResourceInfo]:
            if not nodes:
                return list()

            resources: List[ResourceInfo] = list()
            for node in nodes:
                titles: Dict[str, str] = dict()
                descrs: Dict[str, str] = dict()
                link: Optional[str] = None
                langs: List[str] = list()
                availDataViews: List[DataView] = list()
                availLayers: List[Layer] = list()
                sub: List[ResourceInfo] = list()

                pid = SimpleEndpointDescriptionParser._get_attribute(node, "pid")
                if pid is None:
                    raise SRUConfigException(
                        "Element <ResourceInfo> must have a proper 'pid' attribute"
                    )
                if pid in pids:
                    raise SRUConfigException(
                        f"Another element <Resource> with pid '{pid}' already exists"
                    )
                pids.add(pid)
                LOGGER.debug("Processing resource with pid '%s'", pid)

                for tnode in node.xpath("ed:Title", namespaces={ED_PREFIX: ED_NS}):
                    lang = SimpleEndpointDescriptionParser._get_lang_attribute(tnode)
                    if not lang:
                        raise SRUConfigException(
                            "Element <Title> must have a proper 'xml:lang' attribute"
                        )

                    title = SimpleEndpointDescriptionParser._clean_str(tnode.text)
                    if not title:
                        # NOTE: in java code confusing error message
                        raise SRUConfigException(
                            "Element <Title> must not be non-empty"
                        )
                    if lang in titles:
                        LOGGER.warning("Title with language '%s' already exists", lang)
                    else:
                        LOGGER.debug("title: '%s' '%s'", lang, title)
                        titles[lang] = title
                if titles and LANG_EN not in titles:
                    raise SRUConfigException(
                        "A <Title> with language 'en' is mandatory"
                    )

                for dnode in node.xpath(
                    "ed:Description", namespaces={ED_PREFIX: ED_NS}
                ):
                    lang = SimpleEndpointDescriptionParser._get_lang_attribute(dnode)
                    if not lang:
                        raise SRUConfigException(
                            "Element <Description> must have a proper 'xml:lang' attribute"
                        )

                    descr = SimpleEndpointDescriptionParser._clean_str(dnode.text)

                    if lang in descrs:
                        LOGGER.warning("Title with language '%s' already exists", lang)
                    else:
                        LOGGER.debug("description: '%s' '%s'", lang, descr)
                        # NOTE: skip if None? - java impl would allow nulls
                        if not descr:
                            LOGGER.debug("Skip empty description for lang '%s'", lang)
                        else:
                            descrs[lang] = descr
                if descrs and LANG_EN not in descrs:
                    raise SRUConfigException(
                        "A <Description> with language 'en' is mandatory"
                    )

                for lnode in node.xpath(
                    "ed:LandingPageURI", namespaces={ED_PREFIX: ED_NS}
                ):
                    # TODO: only keep last one? (in java impl)
                    link = SimpleEndpointDescriptionParser._clean_str(lnode.text)

                for lnode in node.xpath(
                    "ed:Languages/ed:Language", namespaces={ED_PREFIX: ED_NS}
                ):
                    val = lnode.text
                    if val:
                        val = val.strip()
                        if not val:
                            val = None

                    # enforce three letter codes
                    if val and len(val) != 3:
                        raise SRUConfigException(
                            "Element <Language> must use ISO-639-3 three letter language codes"
                        )

                    langs.append(val)

                dvnodes = node.xpath(
                    "ed:AvailableDataViews", namespaces={ED_PREFIX: ED_NS}
                )
                if not dvnodes:
                    raise SRUConfigException("Missing element <AvailableDataViews>")
                dvnode = dvnodes[0]
                ref = SimpleEndpointDescriptionParser._get_attribute(dvnode, "ref")
                if not ref:
                    raise SRUConfigException(
                        "Element <AvailableDataViews> must have a 'ref' attribute"
                    )
                refs = ref.split()
                if not refs:
                    raise SRUConfigException(
                        "Attribute 'ref' on element <AvailableDataViews> must contain"
                        " a whitespace seperated list of data view references"
                    )
                for ref in refs:
                    for dataview in supported_DataViews:
                        if ref == dataview.identifier:
                            availDataViews.append(dataview)
                            break
                    else:
                        raise SRUConfigException(
                            f"A data view with identifier '{ref}' was not"
                            " defined in <SupportedDataViews>"
                        )
                if not availDataViews:
                    raise SRUConfigException(
                        f"No available data views were defined for resource"
                        f" with PID '{pid}'"
                    )

                lnodes = node.xpath("ed:AvailableLayers", namespaces={ED_PREFIX: ED_NS})
                if lnodes:
                    lnode = lnodes[0]
                    ref = SimpleEndpointDescriptionParser._get_attribute(lnode, "ref")
                    if not ref:
                        raise SRUConfigException(
                            "Element <AvailableLayers> must have a 'ref' attribute"
                        )
                    refs = ref.split()
                    if not refs:
                        # NOTE: copy-paste error in java
                        raise SRUConfigException(
                            "Attribute 'ref' on element <AvailableLayers> must contain"
                            " a whitespace seperated list of layer references"
                        )
                    for ref in refs:
                        for layer in supported_Layers:
                            if ref == layer.id:
                                availLayers.append(layer)
                                break
                        else:
                            raise SRUConfigException(
                                f"A layer with identifier '{ref}' was not"
                                " defined in <SupportedLayers>"
                            )
                else:
                    if has_ADV_view:
                        LOGGER.debug("No <SupportedLayers> for resource '%s'", pid)

                rnodes = node.xpath(
                    "ed:Resources/ed:Resource", namespaces={ED_PREFIX: ED_NS}
                )
                sub = _parse_resources(rnodes)
                # TODO: None if empty?

                # NOTE: version check in java faulty?
                if availLayers and version <= 1:
                    LOGGER.warning(
                        "Endpoint claims to support FCS 1.0, but includes information"
                        " about <AvailableLayers> for resource with pid '%s'",
                        pid,
                    )

                resources.append(
                    ResourceInfo(
                        pid=pid,
                        title=titles,
                        description=descrs,
                        landing_page_uri=link,
                        languages=langs,
                        available_DataViews=availDataViews,
                        available_Layers=availLayers,
                        sub_Resources=sub,
                    )
                )

            return resources

        nodes = doc.xpath(
            "//ed:EndpointDescription/ed:Resources/ed:Resource",
            namespaces={ED_PREFIX: ED_NS},
        )
        resources: List[ResourceInfo] = _parse_resources(nodes)

        return resources

    # ----------------------------------------------------

    @staticmethod
    def _get_attribute(el: etree._Element, localname: str) -> Optional[str]:
        val: Optional[str] = el.get(localname)
        if val:
            val = val.strip()
            if not val.isspace():
                return val
        return None

    @staticmethod
    def _get_lang_attribute(el: etree._Element) -> Optional[str]:
        name = etree.QName(XML_NS_URI, "lang")
        return SimpleEndpointDescriptionParser._get_attribute(el, name)

    @staticmethod
    def _clean_str(val: str) -> Optional[str]:
        if val:
            val = val.strip()
            val = " ".join(val.split())
            if val:
                return val
        return None

    @staticmethod
    def _dump_ResourceInfo(ris: List[ResourceInfo], depth: int):
        pfx = "--" * depth
        for ri in ris:
            sris = ri.sub_Resources
            LOGGER.debug("%s %s (level=%s)", pfx, ri.pid, depth)
            if sris:
                SimpleEndpointDescriptionParser._dump_ResourceInfo(sris, depth + 1)

    @staticmethod
    def _check_legacy_mode(doc: etree._ElementTree, url: Union[str, TextIO]):
        try:
            if isinstance(url, io.TextIOWrapper) and url.name:
                url = url.name
        except Exception:
            pass
        root: etree.Element = doc.getroot()
        if root is None:
            raise SRUConfigException("Error retrieving root element")
        ns = root.xpath("namespace-uri()")
        if not ns:
            raise SRUConfigException(
                f"No namespace URI was detected for resource info catalog file '{url}'!"
            )
        if ns == RI_NS_LEGACY:
            LOGGER.error(
                f"Detected out-dated resource info catalog file '{url}'."
                "Please update to the current version."
            )
            raise SRUConfigException(f"unsupport file format: {ns}")
        if ns != ED_NS:
            LOGGER.error(
                f"Detected unsupported resource info catalog file '{url}'"
                f" with namespace '{ns}'."
            )
            raise SRUConfigException(f"unsupport file format: {ns}")


# ---------------------------------------------------------------------------



import warnings
from dataclasses import dataclass
from enum import Enum
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union
from xml.sax.handler import ContentHandler

from clarin.sru.xml.writer import XMLStreamWriterHelper

from clarin.sru.fcs.constants import FCS_NS
from clarin.sru.fcs.constants import FCS_PREFIX
from clarin.sru.fcs.constants import FCSDataViewNamespaces

# ---------------------------------------------------------------------------
# basic/HITS/KWIC dataview


class FCSRecordXMLStreamWriter:
    """This class provides several helper methods for writing records
    in the CLARIN-FCS record schema. These methods **do not** cover
    the full spectrum of all variations of records that are permitted
    by the CLARIN-FCS specification.

    See also:
        - CLARIN FCS specification, section "Operation searchRetrieve"
    """

    @staticmethod
    def startResource(
        writer: ContentHandler, pid: Optional[str] = None, ref: Optional[str] = None
    ) -> None:
        """Write the start of a resource (i.e. the ``<Resource>``
        element). Calls to this method need to be balanced with calls
        to the `endResource` method.

        Args:
            writer: the `xml.sax.handler.ContentHandler` to use
            pid: the persistent identifier of this resource or ``None``, if not applicable. Defaults to None.
            ref: the reference of this resource or ``None``, if not applicable. Defaults to None.
        """
        if writer is None:
            raise TypeError("writer is None")

        attrs = dict()
        if pid and not pid.isspace():
            attrs[(None, "pid")] = pid
        if ref and not ref.isspace():
            attrs[(None, "ref")] = ref

        writer.startPrefixMapping(FCS_PREFIX, FCS_NS)
        writer.startElementNS((FCS_NS, "Resource"), None, attrs)

    @staticmethod
    def endResource(writer: ContentHandler) -> None:
        """Write the end of a resource (i.e. the ``</Resource>``
        element). Calls to this method need to be balanced with calls
        to the `startResource` method.

        Args:
            writer: the `xml.sax.handler.ContentHandler` to use
        """
        if writer is None:
            raise TypeError("writer is None")

        writer.endElementNS((FCS_NS, "Resource"), None)
        writer.endPrefixMapping(FCS_PREFIX)

    @staticmethod
    def startResourceFragment(
        writer: ContentHandler, pid: Optional[str] = None, ref: Optional[str] = None
    ) -> None:
        """Write the start of a resource fragment (i.e. the
        ``<ResourceFragment>`` element). Calls to this method need to
        be balanced with calls to the `endResourceFragment` method.

        Args:
            writer: the `xml.sax.handler.ContentHandler` to use
            pid: the persistent identifier of this resource or ``None``, if not applicable. Defaults to None.
            ref: the reference of this resource or ``None``, if not applicable. Defaults to None.
        """
        if writer is None:
            raise TypeError("writer is None")

        attrs = dict()
        if pid and not pid.isspace():
            attrs[(None, "pid")] = pid
        if ref and not ref.isspace():
            attrs[(None, "ref")] = ref

        writer.startElementNS((FCS_NS, "ResourceFragment"), None, attrs)

    @staticmethod
    def endResourceFragment(writer: ContentHandler) -> None:
        """Write the end of a resource fragment (i.e. the
        ``</ResourceFragment>`` element). Calls to this method need
        to be balanced with calls to the `startResourceFragment`
        method.

        Args:
            writer: the `xml.sax.handler.ContentHandler` to use
        """
        if writer is None:
            raise TypeError("writer is None")

        writer.endElementNS((FCS_NS, "ResourceFragment"), None)

    @staticmethod
    def startDataView(writer: ContentHandler, mimetype: str) -> None:
        """Write the start of a data view (i.e. the ``<DataView>``
        element). Calls to this method need to be balanced with calls
        to the `endDataView` method.

        Args:
            writer: the `xml.sax.handler.ContentHandler` to use
            mimetype: the MIME type of this data view applicable
        """
        if writer is None:
            raise TypeError("writer is None")
        if mimetype is None:
            raise TypeError("mimetype is None")
        elif mimetype is None:
            raise ValueError("mimetype is empty")

        writer.startElementNS((FCS_NS, "DataView"), None, {(None, "type"): mimetype})

    @staticmethod
    def endDataView(writer: ContentHandler) -> None:
        """Write the end of a data view (i.e. the ``</DataView>``
        element). Calls to this method need to be balanced with calls
        to the `startDataView` method.

        Args:
            writer: the `xml.sax.handler.ContentHandler` to use
        """
        if writer is None:
            raise TypeError("writer is None")

        writer.endElementNS((FCS_NS, "DataView"), None)

    # ----------------------------------------------------

    @staticmethod
    def writeKWICDataView(
        writer: ContentHandler, left: Optional[str], keyword: str, right: Optional[str]
    ) -> None:
        """[Deprecated] Use the HITS data view instead!
        Convince method to write a KWIC data view. It automatically
        performs the calls to `startDataView` and `endDataView`.

        Args:
            writer: the `xml.sax.handler.ContentHandler` to use
            left: the left context of the KWIC or ``None`` if not applicable
            keyword: the keyword of the KWIC
            right: the right context of the KWIC or ``None`` if not applicable
        """
        warnings.warn(
            "'writeKWICDataView' is deprecated. Use the HITS data view instead!",
            category=DeprecationWarning,
            stacklevel=2,
        )

        if writer is None:
            raise TypeError("writer is None")
        if keyword is None:
            raise TypeError("keyword is None")

        writer = XMLStreamWriterHelper(writer)
        ns = FCSDataViewNamespaces.KWIC

        FCSRecordXMLStreamWriter.startDataView(writer, ns.mimetype)

        # actual "kwic" data view
        with writer.prefix(ns.prefix, ns.namespace), writer.element(
            "kwic", ns.namespace
        ):
            with writer.element("c", ns.namespace, attrs={"type": "left"}):
                if left:
                    writer.characters(left)

            with writer.element("kw", ns.namespace):
                writer.characters(keyword)

            with writer.element("c", ns.namespace, {"type": "right"}):
                if right:
                    writer.characters(right)

        FCSRecordXMLStreamWriter.endDataView(writer)

    # ----------------------------------------------------

    @staticmethod
    def writeSingleHitHitsDataView(
        writer: ContentHandler, left: Optional[str], hit: str, right: Optional[str]
    ) -> None:
        """Convince method to write a simple HITS data view. It
        automatically performs the calls to `startDataView` and
        `endDataView`.

        Args:
            writer: the `xml.sax.handler.ContentHandler` to use
            left: the left context of the hit or ``None`` if not applicable
            hit: the actual hit, that will be highlighted
            right: the right context of the hit or ``None`` if not applicable
        """
        if writer is None:
            raise TypeError("writer is None")
        if hit is None:
            raise TypeError("hit is None")

        writer = XMLStreamWriterHelper(writer)
        ns = FCSDataViewNamespaces.HITS

        FCSRecordXMLStreamWriter.startDataView(writer, ns.mimetype)

        # actual "hits" data view
        with writer.prefix(ns.prefix, ns.namespace), writer.element(
            "Result", ns.namespace
        ):
            if left and not left.isspace():
                writer.characters(left)
            with writer.element("Hit", ns.namespace):
                writer.characters(hit)
            if right and not right.isspace():
                writer.characters(right)

        FCSRecordXMLStreamWriter.endDataView(writer)

    @staticmethod
    def writeHitsDataView(
        writer: ContentHandler,
        text: str,
        hits: List[Tuple[int, int]],
        second_is_length: bool,
    ) -> None:
        """Convince method to write a simple HITS data view. It
        automatically performs the calls to `startDataView` and
        `endDataView`.

        Args:
            writer: the `xml.sax.handler.ContentHandler` to use
            text: the text content of the hit
            hits: a list containing tuples (of start offsets, end offset/length) for the hit markers in the ``text`` content
            second_is_length: if ``True`` the second element of each tuple in this ``hits`` array is interpreted as an length; if ``False`` it is interpreted as an end-offset
        """
        if writer is None:
            raise TypeError("writer is None")
        if text is None:
            raise TypeError("text is None")
        if hits is None:
            raise TypeError("hits is None")
        if not hits or not all(isinstance(e, tuple) and len(e) == 2 for e in hits):
            raise ValueError("hits is empty or not all elements are 2-tuples")

        writer = XMLStreamWriterHelper(writer)
        ns = FCSDataViewNamespaces.HITS

        FCSRecordXMLStreamWriter.startDataView(writer, ns.mimetype)

        # actual "hits" data view
        with writer.prefix(ns.prefix, ns.namespace), writer.element(
            "Result", ns.namespace
        ):
            pos = 0
            for start, end in hits:
                if start < 0 or start > len(text):
                    raise ValueError(f"start index out of bounds: {start=}")
                if second_is_length:
                    if end < 1:
                        raise ValueError(f"length must be larger than 0: length={end}")
                    end += start
                if start >= end:
                    raise ValueError(
                        f"end offset must be larger then start offset: {start=}, {end=}"
                    )

                if start > pos:
                    writer.characters(text[pos:start])

                writer.elementcontent("Hit", text[start:end], ns.namespace)
                pos = end

            if pos < len(text) - 1:
                writer.characters(text[pos:])

        FCSRecordXMLStreamWriter.endDataView(writer)

    # ----------------------------------------------------

    @staticmethod
    def writeResourceWithSingleHitHitsDataView(
        writer: ContentHandler,
        pid: Optional[str],
        ref: Optional[str],
        left: Optional[str],
        hit: str,
        right: Optional[str],
    ):
        """Convince method to write a simple HITS data view. It
        automatically performs the calls to `startResource` and
        `endResource`.

        The following code (arguments omitted) would accomplish the
        same result::

            ...
            FCSRecordXMLStreamWriter.startResource(...)
            FCSRecordXMLStreamWriter.writeSingleHitHitsDataView(...)
            FCSRecordXMLStreamWriter.endResource(...)
            ...

        Args:
            writer: the `xml.sax.handler.ContentHandler` to use
            pid: the persistent identifier of this resource or ``None``, if not applicable.
            ref: the reference of this resource or ``None``, if not applicable.
            left: the left context of the hit or ``None`` if not applicable
            hit: the actual hit, that will be highlighted
            right: the right context of the hit or ``None`` if not applicable

        Raises:
            TypeError: if writer is None
        """
        if writer is None:
            raise TypeError("writer is None")

        FCSRecordXMLStreamWriter.startResource(writer, pid=pid, ref=ref)
        FCSRecordXMLStreamWriter.writeSingleHitHitsDataView(
            writer, left=left, hit=hit, right=right
        )
        FCSRecordXMLStreamWriter.endResource(writer)

    @staticmethod
    def writeResourceWithHitsDataView(
        writer: ContentHandler,
        pid: Optional[str],
        ref: Optional[str],
        text: str,
        hits: List[Tuple[int, int]],
        second_is_length: bool,
    ):
        """Convince method to write a simple HITS data view. It
        automatically performs the calls to `startResource` and
        `endResource`.

        The following code (arguments omitted) would accomplish the
        same result::

            ...
            FCSRecordXMLStreamWriter.startResource(...)
            FCSRecordXMLStreamWriter.writeHitsDataView(...)
            FCSRecordXMLStreamWriter.endResource(...)
            ...

        Args:
            writer: the `xml.sax.handler.ContentHandler` to use
            pid: the persistent identifier of this resource or ``None``, if not applicable.
            ref: the reference of this resource or ``None``, if not applicable.
            text: the text content of the hit
            hits: a list containing tuples (of start offsets, end offset/length) for the hit markers in the ``text`` content
            second_is_length: if ``True`` the second element of each tuple in this ``hits`` array is interpreted as an length; if ``False`` it is interpreted as an end-offset

        Raises:
            TypeError: if writer is None
        """
        if writer is None:
            raise TypeError("writer is None")

        FCSRecordXMLStreamWriter.startResource(writer, pid=pid, ref=ref)
        FCSRecordXMLStreamWriter.writeHitsDataView(
            writer, text=text, hits=hits, second_is_length=second_is_length
        )
        FCSRecordXMLStreamWriter.endResource(writer)

    @staticmethod
    def writeResourceWithKWICDataView(
        writer: ContentHandler,
        pid: Optional[str],
        ref: Optional[str],
        left: Optional[str],
        keyword: str,
        right: Optional[str],
    ) -> None:
        """[Deprecated]
        Convince method for writing a record with a KWIC data view.

        The following code (arguments omitted) would accomplish the
        same result::

            ...
            FCSRecordXMLStreamWriter.startResource(...)
            FCSRecordXMLStreamWriter.writeKWICDataView(...)
            FCSRecordXMLStreamWriter.endResource(...)
            ...

        Args:
            writer: the `xml.sax.handler.ContentHandler` to use
            pid: the persistent identifier of this resource or ``None``, if not applicable.
            ref: the reference of this resource or ``None``, if not applicable.
            left: the left context of the KWIC or ``None`` if not applicable
            keyword: the keyword of the KWIC
            right: the right context of the KWIC or ``None`` if not applicable

        Note:
            Only use, if you want compatability to legacy FCS applications.
        """
        if writer is None:
            raise TypeError("writer is None")

        FCSRecordXMLStreamWriter.startResource(writer, pid=pid, ref=ref)
        FCSRecordXMLStreamWriter.writeKWICDataView(
            writer, left=left, keyword=keyword, right=right
        )
        FCSRecordXMLStreamWriter.endResource(writer)

    @staticmethod
    def writeResourceWithHitsDataViewLegacy(
        writer: ContentHandler,
        pid: Optional[str],
        ref: Optional[str],
        left: Optional[str],
        hit: str,
        right: Optional[str],
    ) -> None:
        """[Deprecated]
        Convince method for writing a record with a HITS and a
        KWIC data view. This method is intended for applications that
        want ensure computability to legacy CLARIN-FCS clients.

        The following code (arguments omitted) would accomplish the
        same result::

            ...
            FCSRecordXMLStreamWriter.startResource(...)
            FCSRecordXMLStreamWriter.writeSingleHitHitsDataView(...)
            FCSRecordXMLStreamWriter.writeKWICDataView(...)
            FCSRecordXMLStreamWriter.endResource(...)
            ...

        Args:
            writer: the `xml.sax.handler.ContentHandler` to use
            pid: the persistent identifier of this resource or ``None``, if not applicable.
            ref: the reference of this resource or ``None``, if not applicable.
            left: the left context of the hit or ``None`` if not applicable
            hit: the actual hit, that will be highlighted
            right: the right context of the hit or ``None`` if not applicable

        Note:
            Only use, if you want compatability to legacy FCS applications.
        """
        if writer is None:
            raise TypeError("writer is None")

        FCSRecordXMLStreamWriter.startResource(writer, pid=pid, ref=ref)
        FCSRecordXMLStreamWriter.writeSingleHitHitsDataView(
            writer, left=left, hit=hit, right=right
        )
        FCSRecordXMLStreamWriter.writeKWICDataView(
            writer, left=left, keyword=hit, right=right
        )
        FCSRecordXMLStreamWriter.endResource(writer)


# ---------------------------------------------------------------------------
# ADV dataview


INITIAL_SEGMENT_ID = -1
NO_HIGHLIGHT = -1


class SpanOffsetUnit(str, Enum):
    ITEM = "item"
    TIMESTAMP = "timestamp"


@dataclass(eq=True)  # (frozen=True)
class Segment:
    id: Union[str, int]
    start: int
    end: int
    # FIXME: add API to set reference
    ref: Optional[str] = None

    def __post_init__(self):
        # object.__setattr__(self, "id", f"s{self.id:x}")
        if isinstance(self.id, int):
            self.id = f"s{self.id:x}"


@dataclass
class Span:
    segment: Segment
    value: Optional[str]
    altValue: Optional[str]
    highlight: Optional[Union[str, int]]

    def __post_init__(self):
        # object.__setattr__(self, "highlight", f"h{self.highlight:x}")
        if self.highlight is not None and isinstance(self.highlight, int):
            if self.highlight != NO_HIGHLIGHT:
                self.highlight = f"h{self.highlight:x}"
            else:
                self.highlight = None


class AdvancedDataViewWriter:
    """Helper class for serializing Advanced Data Views. It can be
    used for writing more than once, but it is **not thread-save**.
    This helper can also serialize HITS Data Views.
    """

    def __init__(self, unit: SpanOffsetUnit) -> None:
        """[Constructor]

        Args:
            unit: the unit to be used for span offsets

        Raises:
            TypeError: if unit is None
        """
        if unit is None:
            raise TypeError("unit is None")
        self.unit = unit
        """the unit to be used for span offsets"""
        self.next_segment_id = INITIAL_SEGMENT_ID
        self.segments: List[Segment] = list()
        self.layers: Dict[str, List[Span]] = dict()

    def reset(self):
        """Reset the writer for writing a new data view (instance)."""
        self.next_segment_id = INITIAL_SEGMENT_ID

    def addSpan(
        self,
        layer_id: str,
        start: int,
        end: int,
        value: Optional[str] = None,
        altValue: Optional[str] = None,
        highlight: Optional[int] = NO_HIGHLIGHT,
    ):
        """Add a span.

        Args:
            layer_id: the span's layer id
            start: the span's start offset
            end: the span's end offset
            value: the span's content value or ``None``
            altValue: the span's alternate value or ``None``
            highlight: the span's alternate value or ``None``
        """
        if layer_id is None:
            raise TypeError("layer_id is None")
        if start < 0:
            raise ValueError("start < 0")
        if end < start:
            raise ValueError("end < start")

        if highlight is None or highlight <= 0:
            highlight = NO_HIGHLIGHT

        # find segment or create a new one
        segment: Segment
        for seg in self.segments:
            if seg.start == start and seg.end == end:
                segment = seg
                break
        else:
            # if not segment:
            segment = Segment(self.next_segment_id, start, end)
            self.segments.append(segment)
            self.next_segment_id += 1

        # find layer or create a new one
        layer: Optional[List[Span]] = self.layers.get(layer_id, None)
        if layer is None:
            layer = list()
            self.layers[layer_id] = layer

        # sanity check (better overlap check?)
        for span in layer:
            if segment == span.segment:
                # FIXME: better exception!
                raise ValueError("segment already exists in layer")
        layer.append(Span(segment, value, altValue, highlight))

    # ----------------------------------------------------

    def writeAdvancedDataView(self, writer: ContentHandler):
        """Write the Advanced Data View to the output stream.

        Args:
            writer: the `xml.sax.handler.ContentHandler` to use
        """
        if writer is None:
            raise TypeError("writer is None")

        writer = XMLStreamWriterHelper(writer)
        ns = FCSDataViewNamespaces.ADV

        FCSRecordXMLStreamWriter.startDataView(writer, ns.mimetype)

        with writer.prefix(ns.prefix, ns.namespace), writer.element(
            "Advanced", ns.namespace, attrs={"unit": self.unit.value}
        ):
            # segments
            with writer.element("Segments", ns.namespace):
                for segment in self.segments:
                    # FIXME: unit translation (long -> time)
                    attrs = {
                        "id": segment.id,
                        "start": str(segment.start),
                        "end": str(segment.end),
                    }
                    if segment.ref:
                        attrs["ref"] = segment.ref
                    writer.startElementNS((ns.namespace, "Segment"), attrs=attrs)
                    writer.endElementNS((ns.namespace, "Segment"))

            # layers
            with writer.element("Layers", ns.namespace):
                for layer_id, layer in self.layers.items():
                    with writer.element("Layer", ns.namespace, attrs={"id": layer_id}):
                        for span in layer:
                            attrs = {"ref": span.segment.id}
                            if span.highlight is not None:
                                attrs["highlight"] = span.highlight
                            if span.altValue is not None:
                                attrs["alt-value"] = span.altValue
                            with writer.element("Span", ns.namespace, attrs=attrs):
                                if span.value and not span.value.isspace():
                                    writer.characters(span.value)

        FCSRecordXMLStreamWriter.endDataView(writer)

    def writeHitsDataView(self, writer: ContentHandler, layer_id: str):
        """Convenience method to write HITS Data View.

        Args:
            writer: the `xml.sax.handler.ContentHandler` to use
            layer_id: the layer id of the layer to be serialized as HITS Data View
        """
        if writer is None:
            raise TypeError("writer is None")
        if layer_id is None:
            raise TypeError("layer_id is None")

        spans = self.layers.get(layer_id, None)
        if not spans:
            raise KeyError(f"layer with id '{layer_id}' does not exist")

        writer = XMLStreamWriterHelper(writer)
        ns = FCSDataViewNamespaces.HITS

        FCSRecordXMLStreamWriter.startDataView(writer, ns.mimetype)

        with writer.prefix(ns.prefix, ns.namespace), writer.element(
            "Result", ns.namespace
        ):
            need_space = False
            for span in spans:
                if need_space:
                    writer.characters(" ")
                    need_space = False

                if span.highlight:
                    with writer.element("Hit", ns.namespace):
                        writer.characters(span.value)
                    need_space = True
                else:
                    writer.characters(span.value)
                    if span.value and not span.value[-1].isspace():
                        need_space = True

        FCSRecordXMLStreamWriter.endDataView(writer)


# ---------------------------------------------------------------------------
